/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.client.onetrickclient;

import std.algorithm;

import dplug.core;
import dplug.client;

import onetrick.client;
import onetrick.ui;
import onetrick.string;

class OneTrickClient : dplug.client.Client
{
	nothrow:
	@nogc:
		this()
		{
			onInitialize();
			onBuildParameterDescriptions();
			super(); // Call last after we've set everything up!
		}
		~this()
		{
			onExit();
			paramDesc.clearContents(); //Automatically freed, it's a struct
			freeAccumulationBuffers();
			if (threadPool !is null)
				threadPool.destroyFree();
		}
		PresetBank internalPresetBank()
		{
			return presetBank;
		}

	    final override void processAudio(const(float*)[] inputs, float*[]outputs, int frames, TimeInfo timeInfo)
	    {
	        foreach(MidiMessage msg; getNextMidiMessages(frames))
	        {
    		    onMidiInput(msg);
			}
			
	    	onUpdateDspParameters();
	    	
	    	onProcessAudio(inputs, outputs, frames, timeInfo);
	    	
	        /// Get access to the GUI
		    if (CustomGUI gui = cast(CustomGUI) graphicsAcquire())
		    {
		    	onAccessGuiUnsafe(gui);
		        graphicsRelease();
		    }
	    }

		final Parameter findParam(string name)
		{
			foreach(p; params)
			{
				if (StringHelper.equals(p.name, name))
					return p;
			}
			return null;//super.param(0);
		}
		
		override PluginInfo buildPluginInfo()
		{
		    static immutable PluginInfo pluginInfo = parsePluginInfo(import("plugin.json"));
		    return pluginInfo;
		}


	protected:
		Vec!ParamDescription paramDesc = makeVec!ParamDescription();
	    Vec!(float*)dspBufferAccum = makeVec!(float*)();
        ThreadPool threadPool;

		void onInitialize() { }

		void onExit() { }
		
		void onReset(double sampleRate, int maxFrames, int numInputs, int numOutputs) { }
		
    	void onMidiInput(MidiMessage msg) { }

    	void onUpdateDspParameters() { writeDspParameters(); }

    	void onProcessAudio(const(float*)[] inputs, float*[]outputs, int frames, TimeInfo timeInfo) { }

		void onAccessGuiUnsafe(CustomGUI gui) { }

		void onBuildParameterDescriptions() { }

		final void allocateAccumulationBuffers(int count)
		{
			// WARNING: Do not use .resize() here, it doesn't zero-fill.. Manually expand with nulls:
			int increase = count - cast(int)dspBufferAccum.length;
			for (int i=0; i<increase; i++)
				dspBufferAccum.pushBack(cast(float*)null);

			foreach (int i; 0..cast(int)dspBufferAccum.length) //Allocate any added buffers
			{
				if (dspBufferAccum[i] is null) 
					dspBufferAccum[i] = mallocSliceNoInit!float(maxFramesInProcess()).ptr;
			}
		}
		
		final void clearAccumulation(int frames)
		{
			//Clear
			//dspBufferAccum[0..outputs.length][0..frames] = 0;
			foreach (int i; 0..cast(int)dspBufferAccum.length)
				dspBufferAccum[i][0..frames] = 0;
		}
		
		final void addParamDescription(ParamDescription desc)
		{
			paramDesc.pushBack(desc);
		}		
		
		final override Parameter[] buildParameters()
		{
			if (paramDesc.length <= 0)
			{
				paramDesc.clearContents();
			    onBuildParameterDescriptions();
			}

			auto params = makeVec!Parameter();
		    
		    // Add in enum order!
		    int count = cast(int)paramDesc.length;
			foreach (int i; 0..count)
			{
				auto p = paramDesc[i];
				/*
					https://dplug.dpldocs.info/dplug.client.params.html
					BoolParameter
					EnumParameter [undocumented]
					FloatParameter
					GainParameter [-inf to value dB log mapping]
					IntegerParameter
					LinearFloatParameter
					LogFloatParameter [frequency mapping]
					Parameter [base?]
					PowFloatParameter [x^N mapping]
				*/
				switch(p.type)
				{
				case ParamType.Dummy:
					params ~= mallocNew!LinearFloatParameter(i, p.name, p.suffix, 0, 1, 0).nonAutomatable();
					break;
				case ParamType.Integer:
					params ~= mallocNew!IntegerParameter(i, p.name, p.suffix, cast(int)p.faustParam.min, cast(int)p.faustParam.max, cast(int)p.faustParam.initial);
					break;
				case ParamType.Float:
					params ~= mallocNew!LinearFloatParameter(i, p.name, p.suffix, p.faustParam.min, p.faustParam.max, p.faustParam.initial);
					break;
				case ParamType.LogFloat:
					params ~= mallocNew!LogFloatParameter(i, p.name, p.suffix, p.faustParam.min, p.faustParam.max, p.faustParam.initial);
					break;
				case ParamType.Gain:
					params ~= mallocNew!GainParameter(i, p.name, p.faustParam.max, p.faustParam.initial);
					break;
			 	case ParamType.Bool:
					params ~= mallocNew!BoolParameter(i, p.name, p.faustParam.initial != 0);
					break;
				default:
				}
			}
			
			return params.releaseData();
		}
		
		
		final void writeDspParameters()
		{
			foreach (int i; 0..cast(int)paramDesc.length)
			{
				auto param = paramDesc[i];
				auto faustParam = param.faustParam;
				if(!faustParam.val)
					continue;
				switch(param.type)
				{
					case ParamType.Integer:
						*(paramDesc[i].faustParam.val) = cast(float)readParam!int(i);
						break;
					case ParamType.Float:
					case ParamType.LogFloat:
						*(faustParam.val) = readParam!float(i);
						break;
					case ParamType.Bool:
						*(paramDesc[i].faustParam.val) = cast(bool)readParam!bool(i) ? 1 : 0;
						break;
					case ParamType.Gain:
						*(faustParam.val) = max(readParam!float(i), -100); //-INF can screw up Faust!
						break;
					default:
						break;
				}
			}
		}
		
		final void computeDspModules(OneTrickDSP[] dspModules, int frames)
		{
			// Lazy load the threadPool
			int threadCount = cast(int)dspModules.length;
			if (threadPool is null)
			{
				threadPool = mallocNew!ThreadPool(threadCount, threadCount);
			}
			else if (threadPool.numThreads < threadCount)
			{
				threadPool.destroyFree();
				threadPool = mallocNew!ThreadPool(threadCount, threadCount);
			}
			
			void workUnit(int m, int threadIndex) nothrow @nogc
			{
				auto dsp = dspModules[m];
				dsp.compute(frames, null, dsp.TempBuffer);
			}
			threadPool.parallelFor(cast(int)dspModules.length, &workUnit);

		}
		
		final void addBuffers(int frames, float*[] buffer, float*[] add)
		{
			//dest[0..dest.length][0..frames] += src[0..dest.length][0..frames];
			int chanCount = cast(int)min(buffer.length, add.length);
			foreach (int chan; 0..chanCount)
			{
				//for(int i=0; i<frames; i++)
				foreach (int i; 0..frames)
				{
					buffer[chan][i] += add[chan][i];
				}
			}
		}

	private:
	
		final void freeAccumulationBuffers()
		{
			foreach (int i; 0..cast(int)dspBufferAccum.length)
			{
				if (dspBufferAccum[i])
				{
					dspBufferAccum[i].destroyFree();
					dspBufferAccum[i] = null;
				}
			}
		}

}
