/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.customgui;

import dplug.math.box;
import dplug.core.nogc;
import dplug.core.file;
import dplug.graphics;

import dplug.gui.graphics;
import dplug.gui.element;
import dplug.gui.compositor;
public import dplug.gui.sizeconstraints;

import onetrick.ui.helper;

interface ICustomGUIListener
{
nothrow:
@nogc:
	void onPreScript();
	void onPostScript();
}

interface ICustomGUI
{
nothrow:
@nogc:
    void addListener(ICustomGUIListener listener);
    void removeListener(ICustomGUIListener listener);
    void parameterChanged();
}

class CustomGUI : GUIGraphics, ICustomGUI
{
public:
nothrow:
@nogc:

    this(SizeConstraints sizeConstraints, UIFlags flags)
    {
        super(sizeConstraints, flags);
        _listeners = makeVec!ICustomGUIListener();
        _drawRectsBackup = makeVec!box2i(0);
    }
    
    ~this()
    {
//    	_listeners.destroyFree();
//		_drawRectsBackup.destroyFree();
    }
    
    override ICompositor buildCompositor(CompositorCreationContext* context)
    {
		//return mallocNew!SimpleRawCompositor(context);
		return mallocNew!BlankCompositor(context);
		//return mallocNew!RoughCompositor(context);
    }
    
    override void regenerateMipmaps() nothrow @nogc
    {
    	// Do nothing!
    }
    override void compositeGUI(ImageRef!RGBA wfb) nothrow @nogc
    {
    	// Do nothing!
    	
    	// Backup _rectsToDisplayDisjointed
    	_drawRectsBackup.clearContents();
        foreach(rect; _rectsToDisplayDisjointed[])
        	_drawRectsBackup.pushBack(rect);

    	// Prevent Composited->Rendered blitting...
    	_rectsToDisplayDisjointed.clearContents();
    }
    override void redrawElementsRaw() nothrow @nogc
    {
    	// Restore _rectsToDisplayDisjointed
    	_rectsToDisplayDisjointed.clearContents();
        foreach(rect; _drawRectsBackup[])
        	_rectsToDisplayDisjointed.pushBack(rect);
    	_drawRectsBackup.clearContents();
    	
    	super.redrawElementsRaw();
    }

    
    /// Adds a parameter listener.
    void addListener(ICustomGUIListener listener)
    {
        _listeners.pushBack(listener);
    }

    /// Removes a parameter listener.
    void removeListener(ICustomGUIListener listener)
    {
        int index = _listeners.indexOf(listener);
        if (index != -1)
            _listeners.removeAndReplaceByLastElement(index);
    }
    
    void parameterChanged()
    {
    	
    }
    
    void notifyPreScript()
    {
		foreach(listener; _listeners)
			listener.onPreScript();
	}
	
	void notifyPostScript()
	{
		foreach(listener; _listeners)
			listener.onPostScript();
	}
    
private:
	Vec!ICustomGUIListener _listeners;
	Vec!box2i _drawRectsBackup;
}

//* // Nothing is needed because we use onDrawRaw() everywhere
final class BlankCompositor : MultipassCompositor
{
nothrow:
@nogc:
    this(CompositorCreationContext* context)
    {
        super(context);
    }
}

//* // This will only have an effect if we use onDrawPBR(), so it's untested right now...
final class RoughCompositor : MultipassCompositor
{
nothrow:
@nogc:
    this(CompositorCreationContext* context)
    {
        super(context);
        addPass(mallocNew!CopyDiffuseToFramebufferRough(this));
    }

    static class CopyDiffuseToFramebufferRough : CompositorPass
    {
    nothrow:
    @nogc:

        this(MultipassCompositor parent)
        {
            super(parent);
        }

        override void render(int threadIndex, const(box2i) area, CompositorPassBuffers* buffers)
        {
            ImageRef!RGBA outputBuf = *(buffers.outputBuf);
            OwnedImage!RGBA diffuseBuf = buffers.diffuseMap.levels[0];
            for (int j = area.min.y; j < area.max.y; ++j)
            {
                RGBA* wfb_scan = outputBuf.scanline(j).ptr;
                RGBA* diffuseScan = diffuseBuf.scanlinePtr(j);

                // write composited color
                for (int i = area.min.x; i < area.max.x; ++i)
                {
                    RGBA df = diffuseScan[i];
                    float grit = cast(float)(i%3+j%2) * 0.2f;
                    grit = grit * 0.1f + 0.9f;
                    wfb_scan[i] = RGBA(cast(ubyte)(df.r*grit), cast(ubyte)(df.g*grit), cast(ubyte)(df.b*grit), 255);
                }
            }
        }
    }
}
//*/
