/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.knob;
import onetrick.ui.parametereditor;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import std.algorithm;
import std.math;
import std.math.trigonometry;

/// A simple knob widget
final class UIKnob : UIParameterEditor
{
public:
nothrow:
@nogc:

	enum MarkerStyle {
		Circle,
		Line,
	}

	enum IndicatorStyle {
		Continuous,
		Led,
	}
	enum IndicatorOrigin {
		Left,
		Center,
	}

	// TODO: Move these to regular @properties that can setDirty
	@ScriptProperty float margin = 0.02;
	@ScriptProperty float indicatorWidth = 0.1;
	@ScriptProperty float knobRadius = 0.65;
	@ScriptProperty float hoverDuration = 90;
	@ScriptProperty float hoverScale = 1.05;
	@ScriptProperty float knobShadeR1 = 0;
	@ScriptProperty float knobShadeR2 = 1;
	@ScriptProperty float knobShadeR3 = 1;

	@ScriptProperty int	  knobMarkerStyle = 0;
	@ScriptProperty float knobMarkerOffset = 0.7;
	@ScriptProperty float knobMarkerCircleRadius = 0.1;
	@ScriptProperty float knobMarkerWidth = 0.1;
	@ScriptProperty float knobMarkerLength = 0.7;
	
	@ScriptProperty int   indicatorStyle = 0;
	@ScriptProperty float indicatorAngleRange = 300;
	@ScriptProperty int   indicatorLedCount = 25;
	@ScriptProperty int   indicatorOrigin = 0;
	@ScriptProperty bool  indicatorDetent = true;
	@ScriptProperty float indicatorGlow = 0.0f;
	@ScriptProperty float indicatorGlowRadius = 1.0f;
	@ScriptProperty float indicatorClipStart = 1.0f;
	
	@ScriptProperty RGBA backingColor = RGBA(0, 0, 0, 0);
	
	@ScriptProperty RGBA knobColor = RGBA(32, 32, 32, 255);
	@ScriptProperty RGBA knobShadeColor = RGBA(0, 0, 0, 0);
	@ScriptProperty RGBA knobShadeColorHover = RGBA(0, 0, 0, 0);
	
	@ScriptProperty RGBA knobMarkerColor = RGBA(7, 219, 255, 255);
	
	@ScriptProperty RGBA indicatorFillColor = RGBA(7, 219, 255, 255);
	@ScriptProperty RGBA indicatorClipColor = RGBA(255, 0, 0, 255);
	@ScriptProperty RGBA indicatorDetentColor = RGBA(255, 255, 255, 255);
	@ScriptProperty RGBA indicatorColor = RGBA(48, 48, 48, 255);
	
	@ScriptProperty RGBA colorDrag = RGBA(255, 255, 0, 255);
	@ScriptProperty RGBA colorHover = RGBA(255, 255, 255, 192);

	this(UIContext context, Parameter param, const(char)[] labelText=null, Font labelFont=null, LabelMode labelMode = LabelMode.Condensed)
	{
		super(context, param, flagRaw | flagAnimated, labelText, labelFont, labelMode);
	}

	void setDirtyKnob()
	{
		box2i pos = position;
		int W = pos.width;
		int H = pos.height;
		vec2f center = vec2f(W*0.5f, H*0.5f);
		float radius = W * 0.5f * knobRadius;
		radius = max(radius, radius * hoverScale)+1; // We could reduce this further by adjusting for old/new hover
		int x0 = cast(int)clamp(center.x-radius, 0, W);
		int y0 = cast(int)clamp(center.y-radius, 0, H);
		int x1 = cast(int)clamp(center.x+radius, 0, W);
		int y1 = cast(int)clamp(center.y+radius, 0, H);
		setDirty(box2i(x0, y0, x1, y1));
	}

	override void onAnimate(double dt, double time)
	{
		super.onAnimate(dt, time); //Needed to refresh label values

		if (hoverScale==1.0f)
			return;
		float targetHover01 = (isMouseOver || isDragged) ? 1.0f : 0.0f;
		if (targetHover01 != _hover01)
		{
			float hoverDirection = (isMouseOver || isDragged) ? 1.0f : -1.0f;
			float newHover01 = clamp(_hover01 + (hoverDirection * dt*1000.0/hoverDuration), 0.0f, 1.0f);
			if (_hover01 != newHover01)
			{
				_hover01 = newHover01;
				//setDirtyWhole();
				setDirtyKnob();
			}
		}
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		super.onDrawRaw(rawMap, dirtyRects);
		float W = position.width;
		float H = position.height;
		vec2f center = vec2f(W * 0.5f, H * 0.5f);
		float size=min(W,H);
		float halfSize = size / 2.0f;
		
		float indicatorOuterRadius = halfSize;
		float indicatorInnerRadius = halfSize * (1.0f-indicatorWidth);
		//Avoid potential artifacts
		indicatorOuterRadius -= 1 + halfSize*margin;
		indicatorInnerRadius -= 1 + halfSize*margin;
		
		float _knobRadius = halfSize*knobRadius * (1.0-margin);
		
		float normalizedValue = param.getNormalized();

		float indicatorAngleStart = (360f-indicatorAngleRange)/2f+90f;

		float indicatorAngleStartR=indicatorAngleStart*TO_RADIANS;
		float indicatorAngleRangeR=indicatorAngleRange*TO_RADIANS;
		float indicatorAngleEndR = indicatorAngleStartR+indicatorAngleRangeR;
		float indicatorAngleCenterR = indicatorAngleStartR+indicatorAngleRangeR * 0.5f;
		float indicatorAngleCurrentR = indicatorAngleStartR+indicatorAngleRangeR*normalizedValue;
		float indicatorAngleClipR = indicatorAngleStartR+indicatorAngleRangeR*indicatorClipStart;
		
		float ledDistance = (indicatorInnerRadius+indicatorOuterRadius) / 2.0f;
		float ledRadius = (indicatorOuterRadius-indicatorInnerRadius) / 2.0f;

		float knobMarkerOffsetRadius = _knobRadius * knobMarkerOffset;
		float _knobMarkerCircleRadius = _knobRadius * knobMarkerCircleRadius;
		vec2f knobMarkerY = vec2f(cos(indicatorAngleCurrentR), sin(indicatorAngleCurrentR));
		vec2f knobMarkerX = vec2f(-knobMarkerY.y, knobMarkerY.x);
		float knobMarkerHalfWidthPx = knobMarkerWidth / 2.0f * _knobRadius;
		float knobMarkerLengthPx = knobMarkerLength * _knobRadius;
		
		//float knobScale = (isMouseOver || isDragged) ? hoverScale : 1.0f;
		float knobScale = 1.0f + (hoverScale-1.0f)*_hover01;
		_knobRadius *= knobScale;
		_knobMarkerCircleRadius *= knobScale;
		knobMarkerOffsetRadius *= knobScale;
		
		RGBA currentIndicatorFillColor = indicatorFillColor;
		if (isMouseOver) currentIndicatorFillColor = blendColor(colorHover, indicatorFillColor, colorHover.a);
		if (isDragged) currentIndicatorFillColor = blendColor(colorDrag, indicatorFillColor, colorDrag.a);
			
		RGBA currentIndicatorClipColor = indicatorClipColor;
		if (isMouseOver && !isDragged) currentIndicatorClipColor = blendColor(colorHover, indicatorClipColor, colorHover.a);
		
		RGBA currentMarkerColor = knobMarkerColor;
		RGBA currentKnobShadeColor = isMouseOver ? knobShadeColorHover : knobShadeColor;


		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			canvas.initialize(croppedRaw);
			canvas.translate(-dirtyRect.min.x, -dirtyRect.min.y); // Offset so we're using a consistent coordinate system
			vec2f rCenter = center - (cast(vec2f)dirtyRect.min + 0.5); // Half texel offset because it seems needed
			
			// Backing
			if (backingColor.a > 0)
			{
				canvas.fillStyle = backingColor;
				canvas.fillCircle(center, halfSize);
			}

			// Indicators
			switch(cast(IndicatorStyle)indicatorStyle)
			{
				case IndicatorStyle.Continuous:
					switch(cast(IndicatorOrigin)indicatorOrigin)
					{
						case IndicatorOrigin.Left:
							// Indicator Background
							croppedRaw.aaFillSector(rCenter.x, rCenter.y, indicatorInnerRadius, indicatorOuterRadius, indicatorAngleCurrentR, indicatorAngleEndR, indicatorColor);


							// Indicator Clip
							if (indicatorAngleCurrentR > indicatorAngleClipR)
							{
								croppedRaw.aaFillSector(rCenter.x, rCenter.y, indicatorInnerRadius, indicatorOuterRadius, indicatorAngleClipR, indicatorAngleCurrentR, currentIndicatorClipColor);
							}
							// Indicator Filled
							croppedRaw.aaFillSector(rCenter.x, rCenter.y, indicatorInnerRadius, indicatorOuterRadius, indicatorAngleStartR, min(indicatorAngleClipR, indicatorAngleCurrentR), currentIndicatorFillColor);
							break;
						case IndicatorOrigin.Center:
							croppedRaw.aaFillSector(rCenter.x, rCenter.y, indicatorInnerRadius, indicatorOuterRadius, indicatorAngleStartR, min(indicatorAngleCurrentR, indicatorAngleCenterR), indicatorColor);
							croppedRaw.aaFillSector(rCenter.x, rCenter.y, indicatorInnerRadius, indicatorOuterRadius, min(indicatorAngleCurrentR, indicatorAngleCenterR), indicatorAngleCenterR, currentIndicatorFillColor);
							croppedRaw.aaFillSector(rCenter.x, rCenter.y, indicatorInnerRadius, indicatorOuterRadius, indicatorAngleCenterR, max(indicatorAngleCurrentR, indicatorAngleCenterR), currentIndicatorFillColor);
							croppedRaw.aaFillSector(rCenter.x, rCenter.y, indicatorInnerRadius, indicatorOuterRadius, max(indicatorAngleCurrentR, indicatorAngleCenterR), indicatorAngleEndR, indicatorColor);
							break;
						default:
					}
					break;
				case IndicatorStyle.Led:
					float angleRangeR = (indicatorAngleEndR - indicatorAngleStartR);
					float step = 1.0f / (cast(float)indicatorLedCount-0.99f); //leave 1% for final led light
					float detentRangeR = angleRangeR*step*0.1f; //Detent Shouldn't have much tollerance
					for (int i=0; i<indicatorLedCount; i++)
					{
						float ratio = cast(float)i * step;
						float angle = indicatorAngleStartR + angleRangeR*ratio;
						float halfAngle = indicatorAngleStartR + angleRangeR * 0.5f;
						vec2f v = vec2f(cos(angle), sin(angle)) * ledDistance;
						bool ledLit = false;
						bool allowDetent = indicatorDetent;
						switch(cast(IndicatorOrigin)indicatorOrigin)
						{
							case IndicatorOrigin.Left:
								ledLit = (angle < indicatorAngleCurrentR);
								allowDetent = false;
								break;
							case IndicatorOrigin.Center:
								ledLit = (angle >= halfAngle*0.99f && angle <= indicatorAngleCurrentR) || 
											(angle <= halfAngle && angle >= indicatorAngleCurrentR);
								break;
							default:
						}
						RGBA iColor = ledLit ? currentIndicatorFillColor : indicatorColor;
						if (allowDetent)
						{
							bool isDetent = (indicatorLedCount%2==1) && i==((indicatorLedCount-1)/2);
							if (isDetent && (abs(angle-indicatorAngleCurrentR) < detentRangeR))
								iColor = indicatorDetentColor;
						}
						//if (useCanvas) {
							canvas.fillStyle = iColor;
							canvas.fillCircle(center+v, ledRadius);					
						/*} else {
							auto p = rCenter+v;
							croppedRaw.aaSoftCircle(p.x, p.y, 0, ledRadius-1, ledRadius, iColor);
						}*/
						
						// Glow
						if (indicatorGlow > 0f && indicatorGlowRadius > 0f && ledLit)
						{
							auto p = rCenter+v;
							croppedRaw.aaSoftCircle(p.x, p.y, ledRadius,ledRadius,ledRadius*(indicatorGlowRadius+1f), iColor, cast(float)iColor.a/255.0f*indicatorGlow);
						}
					}
					break;
				default:
			}
			// Knob
			{
				canvas.fillStyle = knobColor;
				canvas.fillCircle(center, _knobRadius);
				float r1 = knobShadeR1*_knobRadius;
				float r2 = knobShadeR2*_knobRadius;
				float r3 = knobShadeR3*_knobRadius;
				if (knobShadeR2==knobShadeR3)
					r1 -= 1.0f;
				if (currentKnobShadeColor.a > 0)
					croppedRaw.aaSoftCircle(rCenter.x, rCenter.y, r1,r2,r3, currentKnobShadeColor, cast(float)currentKnobShadeColor.a/255.0f);
			}
			
			switch(cast(MarkerStyle)knobMarkerStyle)
			{
				case MarkerStyle.Circle:
					canvas.fillStyle = currentMarkerColor;
					canvas.fillCircle(center+(knobMarkerY * knobMarkerOffsetRadius), _knobMarkerCircleRadius);
					break;
				case MarkerStyle.Line:
					canvas.fillStyle = currentMarkerColor;
					canvas.beginPath();
					canvas.moveTo(center+(knobMarkerY*(knobMarkerOffsetRadius)) + (knobMarkerX*(knobMarkerHalfWidthPx)));
					canvas.lineTo(center+(knobMarkerY*(knobMarkerOffsetRadius-knobMarkerLengthPx)) + (knobMarkerX*(knobMarkerHalfWidthPx)));
					canvas.lineTo(center+(knobMarkerY*(knobMarkerOffsetRadius-knobMarkerLengthPx)) - (knobMarkerX*(knobMarkerHalfWidthPx)));
					canvas.lineTo(center+(knobMarkerY*(knobMarkerOffsetRadius)) - (knobMarkerX*(knobMarkerHalfWidthPx)));
					canvas.closePath();
					canvas.fill();
					break;
				default:
			}
		}
	}

private:
	//	const float PI = 3.14159f;
	const float TO_RADIANS = 1.0f / 360f * PI * 2.0f;
	
	float _hover01 = 0;

	Canvas canvas;
}
