/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.parametereditor;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import std.algorithm;
import std.math.trigonometry;
import std.math.traits; //isNaN

import core.stdc.stdio;
import core.stdc.string: strlen, strcmp;

import onetrick.ui.label;
import onetrick.string.helper;

/// A simple parameter editor base widget
class UIParameterEditor : UIElement, IParameterListener
{
public:
nothrow:
@nogc:
	enum LabelMode {
		Condensed,
		Expanded,
		OnlyLabel,
		OnlyValue,
		None,

		NoValue=OnlyLabel,
		NoLabel=OnlyValue,
	}

	this(UIContext context, Parameter param, uint flags = flagRaw, const(char)[] labelText=null, Font labelFont=null, LabelMode labelMode=LabelMode.Condensed)
	{
		super(context, flags | flagAnimated);
		
		_param = param;
		_param.addListener(this);
		
		this._labelMode = labelMode;
		
		setCursorWhenDragged(MouseCursor.verticalResize);
		
		// Use param.name()?
		// param.label() is the unit (ie. "Hz")
		if (labelText !is null && labelFont !is null) {
			_label = mallocNew!UILabel(context, labelFont, labelText);
			addChild(_label);
			if (labelMode != LabelMode.NoValue)
			{
				_labelValue = mallocNew!UILabel(context, labelFont, "");
				addChild(_labelValue);
				refreshLabelValue();
			}
		}
		updateLabelVisibility();
	}

	~this()
	{
		_param.removeListener(this);
	}
	
	UIParameterEditor withBoolText(string on, string off)
	{
		boolOnText = on;
		boolOffText = off;
		refreshLabelValue();
		return this;
	}


	override void onAnimate(double dt, double time)
	{
		if(atomicLoad(__parameterChanged))
		{
			atomicStore(__parameterChanged, false);
			refreshLabelValue();
		}
		if (__updateLabelVisibilityRequested)
		{
			__updateLabelVisibilityRequested = false;
			updateLabelVisibility();
		}
		updateLabelStyle(); // Calling setX() on UILabel should only update if changed
	}

	override void reflow()
	{
		if (_label !is null) {
			updateLabelStyle();
			auto p = position;
			auto x=p.min.x;
			auto y=p.min.y;
			auto w=p.size.x;
			auto h=p.size.y;
			auto lw = labelSize.x;
			auto lh = labelSize.y;
			if (lw <= 0)
				lw = w;
			if (lh <= 0)
				lh = max(w,h)*0.2;
			
			auto pos = rectangle(cast(int)(x+((w-lw)/2)), cast(int)(y+h+lh*0.1), cast(int)(lw), cast(int)lh);
			_label.position = pos;
			if (_labelValue !is null)
			{
				if (_labelMode != LabelMode.Condensed && _labelMode != LabelMode.OnlyValue)
				{
					pos = pos.translate(0, cast(int)(lh*1.1f));
				}
				_labelValue.position = pos;
			}
		}
	}

	void drawArc(ref Canvas canvas, float x, float y, float innerRadius, float outerRadius, float startAngle, float endAngle, bool anticlockwise)
	{
		drawArc(canvas, vec2f(x,y), innerRadius, outerRadius, startAngle, endAngle, anticlockwise);
	}
	void drawArc(ref Canvas canvas, vec2f center, float innerRadius, float outerRadius, float startAngle, float endAngle, bool anticlockwise)
	{
		canvas.arc(center, outerRadius, startAngle, endAngle, anticlockwise); 
		canvas.arc(center, innerRadius, endAngle, startAngle, !anticlockwise); 
	}

	void refreshLabelValue() {
		if (_labelValue !is null)
		{
			if (auto p = cast(BoolParameter)_param)
			{
				_labelValue.setText(p.getNormalized()>0.5f ? boolOnText : boolOffText);
			}
			else
			{
				double normalizedValue = _param.getNormalized();
				auto suffix = _param.label();
				char[64] valueText;
				_param.stringFromNormalizedValue(normalizedValue, valueText.ptr, valueText.length);
				
				// Hack to get rid of ugly negative zero values!
				bool forceZero = (strcmp(valueText.ptr, "-0.0") == 0);
				char[64] fullLabelText;

				//snprintf(fullLabelText.ptr, fullLabelText.length, "%s%s", (forceZero ? "0.0".ptr : valueText.ptr), CString(suffix).storage);
				StringHelper.formatString(fullLabelText, "%s%s", (forceZero ? "0.0".ptr : valueText.ptr), CString(suffix).storage);
				
				//_param.stringFromNormalizedValue(normalizedValue, fullLabelText.ptr, 32);
				//_labelValue.setText(fullLabelText); // Sometimes leaves orphaned characters at end if sizes change... oops
				_labelValue.setText(fullLabelText[0..strlen(fullLabelText.ptr)]);
			}
			_labelValue.setDirtyWhole();
		}
	}

	override void onBeginParameterHover(Parameter sender) { }
	override void onEndParameterHover(Parameter sender) { } 
	override void onBeginParameterEdit(Parameter sender)
	{
		setDirtyWhole();
	}
	override void onParameterChanged(Parameter sender)
	{
		atomicStore(__parameterChanged, true);
		setDirtyWhole();
	}
	override void onEndParameterEdit(Parameter sender)
	{
		setDirtyWhole();
	}

	override Click onMouseClick(int x, int y, int button, bool isDoubleClick, MouseState mstate)
	{
		bool shouldResetValue = (isDoubleClick || mstate.altPressed);
		
		if(auto p = cast(BoolParameter)_param) // Just use simple toggle for Booleans
		{
			_param.beginParamEdit();
			p.setFromGUI(shouldResetValue ? p.defaultValue() : !p.value);
			_param.endParamEdit();
			return Click.unhandled;
		}
		
		if (shouldResetValue)
		{
			_param.beginParamEdit();
			if (auto p = cast(FloatParameter)_param)
				p.setFromGUI(p.defaultValue());
			else if(auto p = cast(IntegerParameter)_param)
				p.setFromGUI(p.defaultValue());
			else if(auto p = cast(BoolParameter)_param)
				p.setFromGUI(p.defaultValue());
			else
				assert(false); // Unsupported Parameter

			_param.endParamEdit();
			return Click.unhandled;
		}
		
		return Click.startDrag;
	}

	override void onBeginDrag()
	{
		dragDeltaAccum = 0;
		__updateLabelVisibilityRequested=true;
		_param.beginParamEdit();
	}
	override void onStopDrag()
	{
		__updateLabelVisibilityRequested=true;
		_param.endParamEdit();
		dragDeltaAccum = 0;
	}
	override void onMouseDrag(int x, int y, int dx, int dy, MouseState mstate)
	{
		bool horiz = (cursorWhenDragged == MouseCursor.horizontalResize);
		
		float dragDelta = horiz ? (cast(float)(dx) / _position.width) : (cast(float)(-dy) / _position.height);
		float dragSum = dragDelta+dragDeltaAccum;
		
		float modifier = 1.0f;
		if (cast(BoolParameter)_param)
			modifier *= 50.0f; // Make Boolean much more sensitive
			
		if (mstate.shiftPressed || mstate.ctrlPressed) modifier *= 0.1f;
		
		double oldParamValue = _param.getNormalized();
		double newParamValue = oldParamValue + (dragSum) * modifier * sensitivity;
		if (mstate.altPressed)
			newParamValue = _param.getNormalizedDefault();
		if (isNaN(newParamValue)) // Avoid annoying exceptions while developing the UI...
			newParamValue = 0;
		if (newParamValue < 0) newParamValue = 0;
		if (newParamValue > 1) newParamValue = 1;
		
		if (newParamValue != oldParamValue)
		{
			if ( auto p = cast(FloatParameter)_param)
				p.setFromGUINormalized(newParamValue);
			else if(auto p = cast(IntegerParameter)_param)
				p.setFromGUINormalized(newParamValue);
			else if(auto p = cast(BoolParameter)_param)
				p.setFromGUI(newParamValue > 0.5f);
			else
				assert(false); // Unsupported Parameter
		}
		
		bool hitEnd = (newParamValue >= 1 && dragSum > 0) || (newParamValue <= 0 && dragSum < 0);
		if(!hitEnd && _param.getNormalized() == oldParamValue) // No effect, let's accumulate
			dragDeltaAccum += dragDelta;
		else
			dragDeltaAccum = 0;
	}

	override void onMouseEnter()
	{
		__updateLabelVisibilityRequested=true;
		setDirtyWhole();
	}

	override void onMouseExit()
	{
		__updateLabelVisibilityRequested=true;
		setDirtyWhole();
	}
	
	void updateLabelVisibility()
	{
		if (_labelMode == LabelMode.None)
		{
			if (_label !is null)
				_label.visibility = false;
			if (_labelValue !is null)
				_labelValue.visibility = false;
			return;
		}
		if (_labelMode == LabelMode.Condensed || _labelMode == LabelMode.OnlyValue)
		{
			if (_label !is null && _labelValue !is null)
			{
				bool showValue = isMouseOver || _labelMode == LabelMode.OnlyValue;// || isDragged;
				_label.visibility = !showValue;
				_labelValue.visibility = showValue;
				_label.setDirtyWhole();
				_labelValue.setDirtyWhole();				
			}
		}
	}
	
	@property Parameter param() { return _param; }
	
	@property LabelMode labelMode() { return _labelMode; }
	@property void labelMode(LabelMode value)
	{
		if (value != _labelMode)
		{
			_labelMode = value;
			updateLabelVisibility();
			//reflow();
		}
	}
	
	@property vec2f labelSize() { return _labelSize; }
	@property void labelSize(vec2f value)
	{
		if (value != _labelSize)
		{
			_labelSize = value;
			//reflow();
		}
	}
	@property float labelWidth() { return _labelSize.x; }
	@property void labelWidth(float value)
	{
		if (value != _labelSize.x)
		{
			_labelSize.x = value;
			//reflow();
		}
	}
	@property float labelHeight() { return _labelSize.y; }
	@property void labelHeight(float value)
	{
		if (value != _labelSize.y)
		{
			_labelSize.y = value;
			//reflow();
		}
	}
	
	@property RGBA labelColor() { return _labelColor; }
	@property void labelColor(RGBA value)
	{
		if (value != _labelColor)
		{
			_labelColor = value;
			updateLabelStyle();
		}
	}
	@property RGBA labelBackgroundColor() { return _labelBackgroundColor; }
	@property void labelBackgroundColor(RGBA value)
	{
		if (value != _labelBackgroundColor)
		{
			_labelBackgroundColor = value;
			updateLabelStyle();
		}
	}
	@property RGBA labelValueColor() { return _labelValueColor; }
	@property void labelValueColor(RGBA value)
	{
		if (value != _labelValueColor)
		{
			_labelValueColor = value;
			updateLabelStyle();
		}
	}
	@property RGBA labelValueBackgroundColor() { return _labelValueBackgroundColor; }
	@property void labelValueBackgroundColor(RGBA value)
	{
		if (value != _labelValueBackgroundColor)
		{
			_labelValueBackgroundColor = value;
			updateLabelStyle();
		}
	}

protected:
	float sensitivity = 0.2f;

private:
    shared(bool) __parameterChanged = false;  // access to this is through atomic ops
    bool __updateLabelVisibilityRequested = false; // Needed to avoid changing visibility when unsafe.

	void updateLabelStyle()
	{
		if (_label !is null) {  // Calling setX() on UILabel should only update if changed
			_label.color = _labelColor;
			_label.backgroundColor = _labelBackgroundColor;
		}
		if (_labelValue !is null) {  // Calling setX() on UILabel should only update if changed
			_labelValue.color = _labelValueColor;
			_labelValue.backgroundColor = _labelValueBackgroundColor;
		}
	}
	Parameter _param = null;
	
	LabelMode _labelMode = LabelMode.Condensed;
	vec2f _labelSize = vec2f(0,0);
	
	UILabel _label = null;
	RGBA _labelColor = RGBA(255, 255, 255, 255);
	RGBA _labelBackgroundColor = RGBA(0, 0, 0, 64);

	UILabel _labelValue = null;
	RGBA _labelValueColor = RGBA(255, 255, 255, 255);
	RGBA _labelValueBackgroundColor = RGBA(0, 0, 0, 64);
	
	float dragDeltaAccum = 0; // Used for IntegerParameters and general pixel/value aliasing.
	
	string boolOnText = "on";
	string boolOffText = "off";
	
}
