/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.peakindicator;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import std.algorithm;
import std.math.trigonometry;

import onetrick.ui.activatable;
import onetrick.ui.helper;

/// A base animated peak LED
final class UIPeakIndicator : UIElement, IActivatable, IParameterListener
{ mixin Activatable;
public:
nothrow:
@nogc:

	// Hack: Make this transparent to mouse hovering
	// WARNING: Poor Performance, this redraws as the mouse moves!
	//override void onMouseEnter() { context.setMouseOver(parent); }
	// Alternative: Only works for custom isMouseOver tracking,
	override void onMouseEnter() { parent.onMouseEnter(); }
	override void onMouseExit() { parent.onMouseExit(); }

	@ScriptProperty RGBA color = RGBA(64, 64, 64, 255);
	@ScriptProperty RGBA colorActive = RGBA(128, 128, 128, 255);
	@ScriptProperty RGBA colorPeaking = RGBA(255, 255, 0, 255);
	@ScriptProperty float scale = 1;
	@ScriptProperty float scalePeaking = 1;
	@ScriptProperty bool oneShot = false;
	@ScriptProperty float peakHoldMS = 33;
	@ScriptProperty float peakFallMS = 225;
	this(UIContext context, Icon icon=Icon.None)
	{
		super(context, flagRaw | flagAnimated);
		this.icon = icon;
	}
	~this()
	{
		if(iconParam)
		{
			iconParam.removeListener(this);
		}
	}
	
	UIPeakIndicator withSelectableIcon(immutable(Icon)[] icons, IntegerParameter iconParam)
	{
		this.icons = makeVec!Icon;
		foreach(icon; icons)
			this.icons ~= icon;
		this.iconParam = iconParam;
		if(iconParam)
		{
			iconParam.addListener(this);
			updateParameterIcon();
		}
		return this;
	}

	override void onAnimate(double dt, double time)
	{
		setValue(atomicLoad(__peakUpdateValue));
		atomicStore(__peakUpdateValue, 0.0f);
		
		if (cas(&__parameterChanged, true, false))
        {
			updateParameterIcon();
		}
        
        auto dtms = dt*1000.0;
		if (_peakTimer > 0) {
			_peakTimer = max(0, _peakTimer - dtms);
		} else {
			float _lastPeak = _peak;
			float _min = oneShot ? 0 : _value;
			_peak = max(_min, _peak - (dtms/peakFallMS));
			if (_lastPeak!=_peak)
				setDirtyWhole();
		}
			
	}

	void setValue(float value)
	{
		if(!oneShot || value!=_value)
		{
			_value = value;
			if (_value >= _peak)
			{
				_peak = _value;
				_peakTimer = peakHoldMS;
				setDirtyWhole(); // Make sure we get one refresh right now
			}
		}
	}
	void setValueFromDSP(float value)
	{
		// Note: This may get called multiple times per frame, so only store the highest value.
		if (value > atomicLoad(__peakUpdateValue))
		{
			atomicStore(__peakUpdateValue, value); // Set the new Update's value
		}
		//setValue(value); // Unsafe to call directly
	}
	
	void updateParameterIcon()
	{
		if(icons.length > 0)
		{
			int index = cast(int)(iconParam.value() % icons.length);
			icon = icons[index];
			setDirtyWhole();
		}
	}

	override void onBeginParameterHover(Parameter sender) { }
	override void onEndParameterHover(Parameter sender) { } 
	override void onBeginParameterEdit(Parameter sender)
	{
	}
	override void onParameterChanged(Parameter sender)
	{
		atomicStore(__parameterChanged, true);
	}
	override void onEndParameterEdit(Parameter sender)
	{
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		float W = position.width;
		float H = position.height;

		RGBA color = blendColor(colorPeaking, isActive ? colorActive : color, cast(ubyte)(_peak*255));
		float scale = _peak * (scalePeaking-scale) + scale;
		
		{	// IMPORTANT: Quantize scale to avoid flooding the glyph cache!
			const float quantizeSteps = 50;
			scale = cast(float)(cast(int)(scale * quantizeSteps + 0.5)) / quantizeSteps;
		}
		
		//	color = RGBA(cast(ubyte)(_value*255),cast(ubyte)(_value*255),cast(ubyte)(_value*255),255);
		
		//toggleColor = !toggleColor; color = toggleColor ? RGBA(255,255,255,255) : RGBA(0,0,0,255);
		
		if (color.a == 0)
			return;
		
		if (icon != Icon.None)
		{
			vec2f adj = vec2f(W-W*scale, H-H*scale);
			UIHelper.drawIconRaw(rawMap, dirtyRects, box2f.rectangle(adj.x*0.5f, adj.y*0.5f, W-adj.x, H-adj.y), icon, color);
		}
		else
		{
			foreach(dirtyRect; dirtyRects)
			{
				auto croppedRaw = rawMap.cropImageRef(dirtyRect);
				canvas.initialize(croppedRaw);
				canvas.translate(-dirtyRect.min.x, -dirtyRect.min.y); // Offset so we're using a consistent coordinate system
				canvas.fillStyle = color;
				canvas.fillCircle(W/2, H/2, min(W,H)/2*scale);
			}
		}
	}

private:
	// Atomics:
    shared(bool) __parameterChanged = false;
    shared(float) __peakUpdateValue = 0;
    
    
	Canvas canvas;
	Icon icon = Icon.None;
	Vec!Icon icons;
	IntegerParameter iconParam = null;
	float _value = 0;
	float _peak = 0;
	float _peakTimer = 0;

}
