/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick

	OneTrick is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick.  If not, see <http://www.gnu.org/licenses/>.
*/

module onetrick.ui.slider;
import onetrick.ui.parametereditor;
import onetrick.ui.helper;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import std.algorithm;
import std.math;
import std.math.trigonometry;


/// A simple slider widget
final class UISlider : UIParameterEditor
{
public:
nothrow:
@nogc:

	enum GripStyle {
		Switch,
		Fader,
	}
	enum Orientation {
		Horizontal,
		Vertical,
	}

	// TODO: Move these to regular @properties that can setDirty
	@ScriptProperty float margin = 0.2;

	@ScriptProperty GripStyle gripStyle = GripStyle.Fader;
	@ScriptProperty float gripScale = 0.85;
	@ScriptProperty float gripInnerScale = 0.675;
	@ScriptProperty float gripFaderAspect = 16.0 / 9.0;
	@ScriptProperty float gripMarkerWidth = 0.15;
	@ScriptProperty float gripMarkerHeight = 0.8;
	@ScriptProperty float gripHoverScale = 1.20;
	@ScriptProperty float hoverDuration = 90;
	@ScriptProperty float fillScale = 0.8;
	
	
	@ScriptProperty RGBA backingColor = RGBA(64, 64, 64, 255);
	@ScriptProperty RGBA gripColor = RGBA(240, 240, 240, 255);
	@ScriptProperty RGBA gripColorHover = RGBA(250, 250, 250, 255);
	@ScriptProperty RGBA gripInnerColor = RGBA(0, 0, 0, 64);
	@ScriptProperty RGBA gripInnerColorHover = RGBA(0, 0, 0, 64);
	@ScriptProperty RGBA fillColor = RGBA(0, 0, 0, 0);
	@ScriptProperty RGBA fillColorHover = RGBA(0, 0, 0, 0);

	this(UIContext context, Parameter param, const(char)[] labelText=null, Font labelFont=null, LabelMode labelMode = LabelMode.Condensed)
	{
		super(context, param, flagRaw | flagAnimated, labelText, labelFont, labelMode);
		sensitivity = 0.5; //Sliders should be more sensitive
	}


	override void onAnimate(double dt, double time)
	{
		super.onAnimate(dt, time); //Needed to refresh label values

		if (gripHoverScale==1.0f)
			return;
		float targetHover01 = (isMouseOver || isDragged) ? 1.0f : 0.0f;
		if (targetHover01 != _hover01)
		{
			float hoverDirection = (isMouseOver || isDragged) ? 1.0f : -1.0f;
			float newHover01 = clamp(_hover01 + (hoverDirection * dt*1000.0/hoverDuration), 0.0f, 1.0f);
			if (_hover01 != newHover01)
			{
				_hover01 = newHover01;
				setDirtyWhole();
				//setDirtyGrip(); //TODO: Optimize to only update the Grip
			}
		}
	}
	
	override void reflow()
	{
		super.reflow();
		setCursorWhenDragged(horizontal ? MouseCursor.horizontalResize : MouseCursor.verticalResize);
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		super.onDrawRaw(rawMap, dirtyRects);
		bool horiz = horizontal;
		float minXY = min(position.width, position.height);
		int marginAdjust = cast(int)(minXY*margin);
		auto adjustedPosition = position.shrink(marginAdjust);
		box2f sliderPosition = cast(box2f)adjustedPosition.translate(-position.min);
		float W = sliderPosition.width;
		float H = sliderPosition.height;
		float normalizedValue = param.getNormalized();
		if(!horiz)
			normalizedValue = 1.0f-normalizedValue;
		float gripRadius = (horiz?H:W) * 0.5f;
		float gripBaseRadius = gripRadius * gripScale;
		float gripVisualRadius = gripBaseRadius * (1 + (gripHoverScale-1)*_hover01);
		float gripPosMin = gripRadius;
		float gripPosMax = (horiz?W:H)-gripRadius;
		float gripPosRange = gripPosMax-gripPosMin;
		float gripPosXY = gripPosMin+normalizedValue*gripPosRange;
		vec2f gripPos = (horiz ? vec2f(gripPosXY, H/2) : vec2f(W/2, gripPosXY)) + sliderPosition.min;
		//UIHelper.fillColorRaw(rawMap, dirtyRects, RGBA(255,255,255,64));
		if (horiz)
			UIHelper.drawPillHorizontalRaw(rawMap, dirtyRects, sliderPosition, backingColor);
		else
			UIHelper.drawPillVerticalRaw(rawMap, dirtyRects, sliderPosition, backingColor);

		auto hoverByte = cast(ubyte)(_hover01*255);			
		auto currentFillColor = blendColor(fillColorHover, fillColor, hoverByte);
		auto currentGripColor = blendColor(gripColorHover, gripColor, hoverByte);
		auto currentGripInnerColor = blendColor(gripInnerColorHover, gripInnerColor, hoverByte);
		bool isGripTranslucent = currentGripColor.a < 255;

		
		auto fillPosition = sliderPosition.shrink((minXY-(minXY*fillScale)) * 0.25f);
		auto fillRadius = min(fillPosition.width, fillPosition.height) * 0.5f;
		if (horiz)
		{
			fillPosition.max.x = gripPos.x+fillRadius;
			UIHelper.drawPillHorizontalRaw(rawMap, dirtyRects, fillPosition, currentFillColor, true, isGripTranslucent);
		}
		else
		{
			fillPosition.min.y = gripPos.y-fillRadius;
			UIHelper.drawPillVerticalRaw(rawMap, dirtyRects, fillPosition, currentFillColor, isGripTranslucent, true);
		}
		

		switch(gripStyle)
		{
		case GripStyle.Switch:
			UIHelper.drawCircleRaw(rawMap, dirtyRects, gripPos, gripVisualRadius, currentGripColor);
			if(currentGripInnerColor.a > 0)
				UIHelper.drawCircleRaw(rawMap, dirtyRects, gripPos, gripVisualRadius * gripInnerScale, currentGripInnerColor);
			break;
		case GripStyle.Fader:
			{
				float rxScale = horiz ? 1.0/gripFaderAspect : 1;
				float ryScale = horiz ? 1 : 1.0/gripFaderAspect;
				float rx = gripVisualRadius * rxScale;
				float ry = gripVisualRadius * ryScale;
				float rad = gripVisualRadius * 0.25;
				auto gripPosRect = box2f(gripPos.x-rx,gripPos.y-ry,gripPos.x+rx,gripPos.y+ry);
				UIHelper.drawRoundedRectangleRaw(rawMap, dirtyRects, gripPosRect, currentGripColor, rad);
				if(currentGripInnerColor.a > 0)
				{
					rx *= horiz ? gripMarkerWidth : gripMarkerHeight;
					ry *= horiz ? gripMarkerHeight : gripMarkerWidth;
					gripPosRect = box2f(gripPos.x-rx,gripPos.y-ry,gripPos.x+rx,gripPos.y+ry);
					//rad *= gripInnerScale * 0.5;
					UIHelper.drawRoundedRectangleRaw(rawMap, dirtyRects, gripPosRect, currentGripInnerColor, rad);
				}
			}
			break;
		default:
			break;
		}
	}
	
	@property Orientation orientation() { return (position.width > position.height) ? Orientation.Horizontal : Orientation.Vertical; } 
	@property bool horizontal() { return orientation==Orientation.Horizontal; }
	@property bool vertical() { return orientation==Orientation.Vertical; }

private:
	//	const float PI = 3.14159f;
	const float TO_RADIANS = 1.0f / 360f * PI * 2.0f;
	
	float _hover01 = 0;
	Canvas canvas;
}
